// Author: Pete Kirkham
// tiny quad store benchmark.
//
// test merge sort.

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <iostream>

#include "utils/string_utils.h"

/*

*/

void dump (unsigned int* data, size_t length)
{
    std::cout << '[' << data[0];

    for (size_t index(1); index < length; ++index)
        std::cout << ' ' << data[index];

    std::cout << ']' << std::endl;
}

unsigned int swap (unsigned int* data, size_t i, size_t t)
{
    unsigned int temp(data[i]);
    data[i] = data[t];
    data[t] = temp;
    return temp;
}

#ifdef DUMP_MERGE
void dump_merge (unsigned int* data, size_t start, size_t end, size_t i, size_t l, size_t r)
{
    std::cout << '[';

    for (size_t index(start); index < end; ++index)
    {
        std::cout << (index == i ? '|' : ' ') << data[index];
    }

    std::cout << (end == i ? '|' : ' ');

    std::cout << ']';

    std::cout << " i=" << i;
    std::cout << " l=" << l;
    std::cout << " r=" << r;

    std::cout << std::endl << ' ';

    for (size_t index(start); index <= end; ++index)
    {
        std::cout << ' ' << ((index == l || index == r) ? '^' : ' ');
    }

    std::cout << std::endl;
}
#else
#define dump_merge(data,start,end,i,l,r);
#endif

void merge (unsigned int* data, size_t start, size_t mid, size_t end)
{
    if (end == start + 1)
    {
        return;
    }

    if (end == start + 2)
    {
        if (data[start] > data[start+1])
        {
            swap(data, start, start+1);
        }

        return;
    }

    size_t i(start);
    size_t l(0);
    size_t r(mid);

    unsigned int data_r(data[r]);
    unsigned int data_l;
    unsigned int data_i;

    dump_merge(data, start, end, i, l, r);

    // initial scan
    while (data[i] < data_r)
    {
        ++i;
        dump_merge(data, start, end, i, l, r);
    }

    swap(data, i, r);
    l = mid;
    r = mid + 1;
    ++i;

    data_r = data[r];
    data_l = data[l];

    while (i <= mid)
    {
        dump_merge(data, start, end, i, l, r);

        if (data_l <= data_r)
        {
            data_l = swap(data, i, l);

            if ((l < r-1) && (data[l+1] < data_l))
            {
                ++l;
                data_l = data[l];
            }
        }
        else
        {
            data_r = swap(data, i, r);

            if ((r < end-1) && (data[r+1] < data_r))
            {
                ++r;
                data_r = data[r];
            }
        }

        ++i;
    }


    while (i < end)
    {
        if (i >= l)
        {
            l = i+1;

            data_l = data[l];

            if (l >= r)
            {
                r = l+1;

                if (r >= end)
                {
                    if (i < end-1)
                    {
                        if (data[end-1] < data[end-2])
                        {
                            swap(data, end-1, end-2);
                        }
                    }

                    break;
                }

                data_r = data[r];
            }
        }

        dump_merge(data, start, end, i, l, r);

        data_i = data[i];

        //~ std::cout << "data_i " << data_i << ' ';
        //~ std::cout << "data_l " << data_l << ' ';
        //~ std::cout << "data_r " << data_r << std::endl;

        if ((data_i >= data_l) || (data_i >= data_r))
        {
            //~ std::cout << "((data_i >= data_l) || (data_i >= data_r)) " << std::endl;

            if (data_l <= data_r)
            {
                //~ std::cout << "(data_l <= data_r) " << std::endl;

                data_l = swap(data, i, l);

                if ((l < r-1) && (data[l+1] < data_l))
                {
                    ++l;
                    data_l = data[l];
                }
            }
            else
            {
                data_r = swap(data, i, r);

                if ((r < end-1) && (data[r+1] < data_r))
                {
                    ++r;
                    data_r = data[r];
                }
            }
        }

        ++i;
    }

    dump_merge(data, start, end, i, l, r);
}

inline size_t min (size_t a, size_t b) { return a <= b ? a : b; }

void merge_sort (unsigned int* data, size_t length)
{
    for (size_t step(1); step < length; step <<= 1)
    {
        for (size_t index(0); index < length; index += 2*step)
        {
            merge(data, index, index + step, min(index + 2 * step, length));
        }
    }
}

int main (int narg, char** argvp)
{
    size_t length(narg > 1 ? atoi(argvp[1]) : 12);

    unsigned int* data = new unsigned int[length];

    //~ for (size_t index(0); index < length; ++index)
    //~ {
        //~ data[index] = index % (length/2);//mix(index) % (length/3);
    //~ }

    //~ merge(data, 0, length/2, length);

    for (size_t seed(0); ; ++seed)
    {
        unsigned int value(seed);
        unsigned int sum(0);

        for (size_t index(0); index < length; ++index)
        {
            data[index] = value % length;
            value /= length;
            sum += data[index];
        }

        if ((sum == 0) && (seed != 0)) break;

        dump(data, length);

        merge_sort(data, length);

        dump(data, length);

        std::cout << std::endl;

        for (size_t index(1); index < length; ++index)
        {
            if (data[index] < data[index-1])
            {
                std::cerr << seed << std::endl;
                exit(-1);
            }
        }
    }

    delete [] data;

    return 0;
}
